/*
    ext2prepare.c -- prepares an ext2 fs for an online resize
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This program is to be used in cooperation with Andreas Dilger's
  online resize patch for the linux kernel.
*/

static const char _ext2prepare_c[] = "$Id: ext2prepare.c,v 1.1 1999/07/31 08:22:31 buytenh Exp $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include "ext2.h"

blk_t getfreezeroedblock(struct ext2_fs *fs)
{
	int i;

	for (i=0;i<fs->numgroups;i++)
	{
		blk_t offset;
		blk_t j;

		offset = i * fs->sb.s_blocks_per_group + fs->sb.s_first_data_block;
		for (j=fs->itoffset+fs->inodeblocks;j<fs->sb.s_blocks_per_group;j++)
			if (!ext2_get_block_state(fs, offset+j))
			{
				struct ext2_buffer_head *bh;

				ext2_set_block_state(fs, offset+j, 1, 1);
				bh = ext2_bcreate(fs, offset+j);
				bh->dirty = 1;
				ext2_brelse(bh, 0);

				return offset+j;
			}
	}

	return 0;
}

int putblk(struct ext2_fs *fs, struct ext2_inode *inode, blk_t block)
{
	struct ext2_buffer_head *bh;
	__u32			*udata;
	int			 i;
	int			 usperblock;

	usperblock = fs->blocksize >> 2;

	for (i=0;i<12;i++)
		if (!inode->i_block[i])
		{
			inode->i_block[i] = block;
			inode->i_blocks++;
			return 1;
		}

	if (!inode->i_block[EXT2_IND_BLOCK])
	{
		if (!(inode->i_block[EXT2_IND_BLOCK] = getfreezeroedblock(fs)))
			return 0;

		inode->i_blocks++;
	}
	bh = ext2_bread(fs, inode->i_block[EXT2_IND_BLOCK]);
	udata = (__u32 *)bh->data;

	for (i=0;i<usperblock;i++)
		if (!udata[i])
		{
			bh->dirty = 1;
			udata[i] = block;
			inode->i_blocks++;
			return 1;
		}

	ext2_brelse(bh, 0);

#if 0
	index -= usperblock;

	if (index < usperblock * usperblock)
	{
		struct ext2_buffer_head *bh0;
		struct ext2_buffer_head *bh1;
		int num0;
		int num1;
		__u32 *udata;

		num0 = index / usperblock;
		num1 = index % usperblock;

		if (!inode->i_block[EXT2_DIND_BLOCK])
		{
			if (!(inode->i_block[EXT2_DIND_BLOCK] = getfreezeroedblock(fs)))
				return 0;

			inode->i_blocks++;
		}

		bh0 = ext2_bread(fs, inode->i_block[EXT2_DIND_BLOCK]);
		udata = (__u32 *)bh0->data;

		if (!udata[num0])
		{
			if (!(udata[num0] = getfreezeroedblock(fs)))
			{
				ext2_brelse(bh0, 0);
				return 0;
			}

			bh0->dirty = 1;
			inode->i_blocks++;
		}

		bh1 = ext2_bread(fs, udata[num0]);
		bh1->dirty = 1;
		udata = (__u32 *)bh1->data;
		udata[num1] = block;
		ext2_brelse(bh1, 0);
		ext2_brelse(bh0, 0);
		inode->i_blocks++;

		return 1;
	}
#endif

	return 0;
}

int create_inode_007(struct ext2_fs *fs)
{
	int			diff;
	int			i;
	struct ext2_inode	inode;

	ext2_read_inode(fs, 7, &inode);
	if (inode.i_mode)
	{
		fprintf(stderr, "inode 007 already allocated!\n");
		return 0;
	}

	diff = fs->itoffset - fs->gdblocks - 2;

	inode.i_mode = 0100400;
	inode.i_size = (fs->numgroups * diff) << fs->logsize;
	inode.i_uid = inode.i_gid = 0;
	inode.i_atime = inode.i_ctime =	inode.i_mtime =	inode.i_dtime = 0;
	inode.i_links_count = 1;
	inode.i_blocks = 0;
	inode.i_flags = 0;
	inode.osd1.linux1.l_i_reserved1 = 0;
	inode.i_block[0] = 0;
	inode.i_block[1] = 0;
	inode.i_block[2] = 0;
	inode.i_block[3] = 0;
	inode.i_block[4] = 0;
	inode.i_block[5] = 0;
	inode.i_block[6] = 0;
	inode.i_block[7] = 0;
	inode.i_block[8] = 0;
	inode.i_block[9] = 0;
	inode.i_block[10] = 0;
	inode.i_block[11] = 0;
	inode.i_block[12] = 0;
	inode.i_block[13] = 0;
	inode.i_block[14] = 0;
	inode.i_version = 0;
	inode.i_file_acl = 0;
	inode.i_dir_acl = 0;
	inode.i_faddr = 0;
	inode.osd2.linux2.l_i_frag = 0;
	inode.osd2.linux2.l_i_fsize = 0;
	inode.osd2.linux2.i_pad1 = 0;
	inode.osd2.linux2.l_i_reserved2[0] = 0;
	inode.osd2.linux2.l_i_reserved2[1] = 0;

	for (i=0;i<fs->numgroups;i++)
	{
		if (!putblk(fs, &inode, 0))
			return 0;
	}

	ext2_write_inode(fs, 14, &inode);

	return 1;
}


int ext2_prepare_fs(struct ext2_fs *fs, blk_t maxnewsize)
{
	int status;

#if PROGRESS
	fprintf(stderr, "ext2_prepare_fs\n");
#endif

	if (maxnewsize <= fs->sb.s_blocks_count)
		return 1;

	if ((fs->relocator_pool = (unsigned char *)malloc(ext2_relocator_pool_size << 10)) == NULL)
	{
		fprintf(stderr, "ext2_prepare_fs: error allocating relocator pool!\n");
		return 0;
	}
	fs->relocator_pool_end = fs->relocator_pool + (ext2_relocator_pool_size << 10);

	status = ext2_block_relocate(fs, maxnewsize);

	if (status)
		status = ext2_metadata_push(fs, maxnewsize);

#if 0
	if (status)
		status = create_inode_007(fs);
#endif

	free(fs->relocator_pool);
	fs->relocator_pool = NULL;
	fs->relocator_pool_end = NULL;

	return status;
}

int main(int argc, char *argv[])
{
	struct ext2_dev_handle *handle;
	struct ext2_fs	       *fs;
	int		        newgroups;
	blk_t		        newsize;

	if (argc != 3)
	{
		fprintf(stderr, "syntax: %s <device> <max. new size>\n", argv[0]);
		exit(-1);
	}

	handle = ext2_make_dev_handle_from_file(argv[1]);
	if (handle == NULL)
	{
		fprintf(stderr, "cant open %s\n", argv[1]);
		return 1;
	}

	sscanf(argv[2], "%i", &newsize);

	newgroups = (howmany(newsize, 4096) + 31) & ~0x1f;
	if (ext2_max_groups < newgroups)
		ext2_max_groups = newgroups;

	if ((fs = ext2_open(handle)) == NULL)
	{
		fprintf(stderr, "cant open %s\n", argv[1]);
		exit(-1);
	}

	if (!ext2_prepare_fs(fs, newsize))
		return 1;

	ext2_close(fs);

	return 0;
}

